"""
*******************************************************************************
*   Ledger Blue
*   (c) 2016 Ledger
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*      http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
********************************************************************************
"""

from .comm import getDongle
import binascii
import argparse


def auto_int(x):
    return int(x, 0)

parser = argparse.ArgumentParser()
parser.add_argument(
    "--key", help="Reference of the endorsement key to setup (1 or 2)", type=auto_int)
parser.add_argument(
    "--certificate", help="Certificate to store if finalizing the endorsement (hex encoded)")
parser.add_argument("--apdu", help="Display APDU log", action='store_true')

args = parser.parse_args()

if args.key == None:
    raise Exception("Missing endorsement key reference")
if args.key != 1 and args.key != 2:
    raise Exception("Invalid endorsement key reference")

dongle = getDongle(args.apdu)
if args.certificate == None:
    apdu = bytearray([0xe0, 0xC0, args.key, 0x00, 0x00])
    response = dongle.exchange(apdu)
    print("Public key " + str(response[0:65]).encode('hex'))
    print("Certificate " + str(response[65:]).encode('hex'))
else:
    certificate = bytearray.fromhex(args.certificate)
    apdu = bytearray([0xe0, 0xC2, 0x00, 0x00, len(certificate)]) + certificate
    dongle.exchange(apdu)
    print("Endorsement setup finalized")



"""
*******************************************************************************
*   Ledger Blue
*   (c) 2016 Ledger
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*      http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
********************************************************************************
"""

from .ecWrapper import PrivateKey, PublicKey
from .comm import getDongle
from .commException import CommException
from .hexParser import IntelHexParser
from .hexLoader import HexLoader
import argparse
import struct
import os
import binascii

def auto_int(x):
    return int(x, 0)

def getDeployedSecretV2(dongle, masterPrivate, targetid):
        testMaster = PrivateKey(bytes(masterPrivate))
        testMasterPublic = bytearray(testMaster.pubkey.serialize(compressed=False))
        targetid = bytearray(struct.pack('>I', targetid))

        # identify
        apdu = bytearray([0xe0, 0x04, 0x00, 0x00]) + bytearray([len(targetid)]) + targetid
        dongle.exchange(apdu)

        # walk the chain 
        nonce = os.urandom(8)
        apdu = bytearray([0xe0, 0x50, 0x00, 0x00]) + bytearray([len(nonce)]) + nonce
        auth_info = dongle.exchange(apdu)
        batch_signer_serial = auth_info[0:4]
        deviceNonce = auth_info[4:12]

        # if not found, get another pair
        #if cardKey != testMasterPublic:
        #       raise Exception("Invalid batch public key")

        dataToSign = bytes(bytearray([0x01]) + testMasterPublic)        
        signature = testMaster.ecdsa_sign(bytes(dataToSign))
        signature = testMaster.ecdsa_serialize(signature)
        certificate = bytearray([len(testMasterPublic)]) + testMasterPublic + bytearray([len(signature)]) + signature
        apdu = bytearray([0xE0, 0x51, 0x00, 0x00]) + bytearray([len(certificate)]) + certificate
        dongle.exchange(apdu)
        
        # provide the ephemeral certificate
        ephemeralPrivate = PrivateKey()
        ephemeralPublic = bytearray(ephemeralPrivate.pubkey.serialize(compressed=False))
        dataToSign = bytes(bytearray([0x11]) + nonce + deviceNonce + ephemeralPublic)
        signature = testMaster.ecdsa_sign(bytes(dataToSign))
        signature = testMaster.ecdsa_serialize(signature)
        certificate = bytearray([len(ephemeralPublic)]) + ephemeralPublic + bytearray([len(signature)]) + signature
        apdu = bytearray([0xE0, 0x51, 0x80, 0x00]) + bytearray([len(certificate)]) + certificate
        dongle.exchange(apdu)

        # walk the device certificates to retrieve the public key to use for authentication
        index = 0
        last_pub_key = PublicKey(binascii.unhexlify("0490f5c9d15a0134bb019d2afd0bf297149738459706e7ac5be4abc350a1f818057224fce12ec9a65de18ec34d6e8c24db927835ea1692b14c32e9836a75dad609"), raw=True)
        while True:
                if index == 0:                  
                        certificate = bytearray(dongle.exchange(bytearray.fromhex('E052000000')))
                elif index == 1:
                        certificate = bytearray(dongle.exchange(bytearray.fromhex('E052800000')))
                else:
                                break
                if len(certificate) == 0:
                        break
                offset = 1
                certificateHeader = certificate[offset : offset + certificate[offset-1]]
                offset += certificate[offset-1] + 1
                certificatePublicKey = certificate[offset : offset + certificate[offset-1]]
                offset += certificate[offset-1] + 1
                certificateSignatureArray = certificate[offset : offset + certificate[offset-1]]
                certificateSignature = last_pub_key.ecdsa_deserialize(bytes(certificateSignatureArray))
                # first cert contains a header field which holds the certificate's public key role
                if index == 0:
                        certificateSignedData = bytearray([0x02]) + certificateHeader + certificatePublicKey
                        # Could check if the device certificate is signed by the issuer public key
                # ephemeral key certificate
                else:
                        certificateSignedData = bytearray([0x12]) + deviceNonce + nonce + certificatePublicKey          
                if not last_pub_key.ecdsa_verify(bytes(certificateSignedData), certificateSignature):
                        return None
                last_pub_key = PublicKey(bytes(certificatePublicKey), raw=True)
                index = index + 1

        # Commit device ECDH channel
        dongle.exchange(bytearray.fromhex('E053000000'))
        secret = last_pub_key.ecdh(binascii.unhexlify(ephemeralPrivate.serialize()))
        return secret[0:16]


parser = argparse.ArgumentParser()
parser.add_argument("--targetId", help="Set the chip target ID", type=auto_int)
parser.add_argument("--apdu", help="Display APDU log", action='store_true')

args = parser.parse_args()

if args.targetId == None:
        args.targetId = 0x31000002

privateKey = PrivateKey()
publicKey = str(privateKey.pubkey.serialize(compressed=False))
args.rootPrivateKey = privateKey.serialize()

genuine = False

dongle = getDongle(args.apdu)

secret = getDeployedSecretV2(dongle, bytearray.fromhex(args.rootPrivateKey), args.targetId)
if secret != None:
        loader = HexLoader(dongle, 0xe0, True, secret)
        data = b'\xFF'
        data = loader.encryptAES(data)
        try:
                loader.exchange(loader.cla, 0x00, 0x00, 0x00, data)                             
        except CommException as e:
                genuine = (e.sw == 0x6D00)

if genuine:
        print ("Product is genuine")
else:
        print ("Product is NOT genuine")




"""
*******************************************************************************
*   Ledger Blue
*   (c) 2016 Ledger
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*      http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
********************************************************************************
"""

from .hexParser import IntelHexParser
from .hexParser import IntelHexPrinter
import hashlib
import binascii
import argparse

def auto_int(x):
    return int(x, 0)

parser = argparse.ArgumentParser()
parser.add_argument("--hex", help="Hex file to be hashed")

args = parser.parse_args()

if args.hex == None:
    raise Exception("Missing hex filename to hash")

# parse
parser = IntelHexParser(args.hex)

# prepare data
m = hashlib.sha256()
# consider areas are ordered by ascending address and non-overlaped
for a in parser.getAreas():
    m.update(a.data)
dataToSign = m.digest()

print(dataToSign.encode('hex'))



"""
*******************************************************************************
*   Ledger Blue
*   (c) 2016 Ledger
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*      http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
********************************************************************************
"""

from .ecWrapper import PublicKey
import hashlib
import binascii
import argparse

parser = argparse.ArgumentParser()
parser.add_argument("--key", help="The public endorsement key 1 to verify with (hex encoded)")
parser.add_argument("--codehash", help="The hash of the code associated to the endorsement request (hex encoded)")
parser.add_argument("--message", help="The message associated to the endorsement request (hex encoded)")
parser.add_argument("--signature", help="The signature to verify with (hex encoded)")

args = parser.parse_args()

if args.key == None:
	raise Exception("Missing public key")
if args.codehash == None:
	raise Exception("Missing code hash")
if args.message == None:
	raise Exception("Missing message")
if args.signature == None:
	raise Exception("Missing signature")

# prepare data
m = hashlib.sha256()
m.update(bytes(bytearray.fromhex(args.message)))
m.update(bytes(bytearray.fromhex(args.codehash)))
digest = m.digest()

publicKey = PublicKey(bytes(bytearray.fromhex(args.key)), raw=True)
signature = publicKey.ecdsa_deserialize(bytes(bytearray.fromhex(args.signature)))
if not publicKey.ecdsa_verify(bytes(digest), signature, raw=True):
	raise Exception("Endorsement not verified")

print("Endorsement verified")



"""
*******************************************************************************
*   Ledger Blue
*   (c) 2016 Ledger
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*      http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
********************************************************************************
"""
from .hexParser import IntelHexParser
from .hexLoader import HexLoader
from .comm import getDongle
import argparse

def auto_int(x):
    return int(x, 0)

parser = argparse.ArgumentParser()
parser.add_argument("--targetId", help="Set the chip target ID", type=auto_int)
parser.add_argument("--fileName", help="Set the file name to load")
parser.add_argument("--bootAddr", help="Set the boot address", type=auto_int)
parser.add_argument("--apdu", help="Display APDU log", action='store_true')

args = parser.parse_args()

if args.targetId == None:
	raise Exception("Missing targetId")
if args.fileName == None:
	raise Exception("Missing fileName")

parser = IntelHexParser(args.fileName)
if args.bootAddr == None:
    args.bootAddr = parser.getBootAddr()

dongle = getDongle(args.apdu)

#relative load
loader = HexLoader(dongle, 0xe0, False, None, False)

loader.validateTargetId(args.targetId)
hash = loader.load(0xFF, 0xF0, parser)
loader.run(parser.getAreas(), args.bootAddr)



"""
*******************************************************************************
*   Ledger Blue
*   (c) 2016 Ledger
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*      http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
********************************************************************************
"""

from .ecWrapper import PrivateKey
from .comm import getDongle
from .deployed import getDeployedSecretV1, getDeployedSecretV2
from .hexLoader import HexLoader
import argparse
import binascii
import sys

def auto_int(x):
    return int(x, 0)

parser = argparse.ArgumentParser()
parser.add_argument("--targetId", help="Set the chip target ID", type=auto_int)
parser.add_argument("--appName", help="Set the application name")
parser.add_argument("--rootPrivateKey", help="Set the root private key")
parser.add_argument("--apdu", help="Display APDU log", action='store_true')
parser.add_argument("--deployLegacy", help="Use legacy deployment API", action='store_true')

args = parser.parse_args()

if args.appName == None:
	raise Exception("Missing appName")
    
if (sys.version_info.major == 3):
	args.appName = bytes(args.appName,'ascii')
if (sys.version_info.major == 2):
	args.appName = bytes(args.appName)

if args.targetId == None:
	args.targetId = 0x31000002
if args.rootPrivateKey == None:
	privateKey = PrivateKey()
	publicKey = binascii.hexlify(privateKey.pubkey.serialize(compressed=False))
	print("Generated random root public key : %s" % publicKey)
	args.rootPrivateKey = privateKey.serialize()

dongle = getDongle(args.apdu)

if args.deployLegacy:
	secret = getDeployedSecretV1(dongle, bytearray.fromhex(args.rootPrivateKey), args.targetId)
else:
	secret = getDeployedSecretV2(dongle, bytearray.fromhex(args.rootPrivateKey), args.targetId)
loader = HexLoader(dongle, 0xe0, True, secret)
loader.deleteApp(args.appName)



"""
*******************************************************************************
*   Ledger Blue
*   (c) 2016 Ledger
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*      http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
********************************************************************************
"""

from Crypto.Cipher import AES
import struct
import hashlib
import binascii

LOAD_SEGMENT_CHUNK_HEADER_LENGTH = 3
MIN_PADDING_LENGTH = 1

class HexLoader:
	def __init__(self, card, cla=0xF0, secure=False, key=None, relative=True):
		self.card = card
		self.cla = cla
		self.secure = secure
		self.key = key
		self.iv = b"\x00" * 16
		self.relative = relative

	
		
	def crc16(self, data):
		TABLE_CRC16_CCITT = [
			0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50a5, 0x60c6, 0x70e7,
			0x8108, 0x9129, 0xa14a, 0xb16b, 0xc18c, 0xd1ad, 0xe1ce, 0xf1ef,
			0x1231, 0x0210, 0x3273, 0x2252, 0x52b5, 0x4294, 0x72f7, 0x62d6,
			0x9339, 0x8318, 0xb37b, 0xa35a, 0xd3bd, 0xc39c, 0xf3ff, 0xe3de,
			0x2462, 0x3443, 0x0420, 0x1401, 0x64e6, 0x74c7, 0x44a4, 0x5485,
			0xa56a, 0xb54b, 0x8528, 0x9509, 0xe5ee, 0xf5cf, 0xc5ac, 0xd58d,
			0x3653, 0x2672, 0x1611, 0x0630, 0x76d7, 0x66f6, 0x5695, 0x46b4,
			0xb75b, 0xa77a, 0x9719, 0x8738, 0xf7df, 0xe7fe, 0xd79d, 0xc7bc,
			0x48c4, 0x58e5, 0x6886, 0x78a7, 0x0840, 0x1861, 0x2802, 0x3823,
			0xc9cc, 0xd9ed, 0xe98e, 0xf9af, 0x8948, 0x9969, 0xa90a, 0xb92b,
			0x5af5, 0x4ad4, 0x7ab7, 0x6a96, 0x1a71, 0x0a50, 0x3a33, 0x2a12,
			0xdbfd, 0xcbdc, 0xfbbf, 0xeb9e, 0x9b79, 0x8b58, 0xbb3b, 0xab1a,
			0x6ca6, 0x7c87, 0x4ce4, 0x5cc5, 0x2c22, 0x3c03, 0x0c60, 0x1c41,
			0xedae, 0xfd8f, 0xcdec, 0xddcd, 0xad2a, 0xbd0b, 0x8d68, 0x9d49,
			0x7e97, 0x6eb6, 0x5ed5, 0x4ef4, 0x3e13, 0x2e32, 0x1e51, 0x0e70,
			0xff9f, 0xefbe, 0xdfdd, 0xcffc, 0xbf1b, 0xaf3a, 0x9f59, 0x8f78,
			0x9188, 0x81a9, 0xb1ca, 0xa1eb, 0xd10c, 0xc12d, 0xf14e, 0xe16f,
			0x1080, 0x00a1, 0x30c2, 0x20e3, 0x5004, 0x4025, 0x7046, 0x6067,
			0x83b9, 0x9398, 0xa3fb, 0xb3da, 0xc33d, 0xd31c, 0xe37f, 0xf35e,
			0x02b1, 0x1290, 0x22f3, 0x32d2, 0x4235, 0x5214, 0x6277, 0x7256,
			0xb5ea, 0xa5cb, 0x95a8, 0x8589, 0xf56e, 0xe54f, 0xd52c, 0xc50d,
			0x34e2, 0x24c3, 0x14a0, 0x0481, 0x7466, 0x6447, 0x5424, 0x4405,
			0xa7db, 0xb7fa, 0x8799, 0x97b8, 0xe75f, 0xf77e, 0xc71d, 0xd73c,
			0x26d3, 0x36f2, 0x0691, 0x16b0, 0x6657, 0x7676, 0x4615, 0x5634,
			0xd94c, 0xc96d, 0xf90e, 0xe92f, 0x99c8, 0x89e9, 0xb98a, 0xa9ab,
			0x5844, 0x4865, 0x7806, 0x6827, 0x18c0, 0x08e1, 0x3882, 0x28a3,
			0xcb7d, 0xdb5c, 0xeb3f, 0xfb1e, 0x8bf9, 0x9bd8, 0xabbb, 0xbb9a,
			0x4a75, 0x5a54, 0x6a37, 0x7a16, 0x0af1, 0x1ad0, 0x2ab3, 0x3a92,
			0xfd2e, 0xed0f, 0xdd6c, 0xcd4d, 0xbdaa, 0xad8b, 0x9de8, 0x8dc9,
			0x7c26, 0x6c07, 0x5c64, 0x4c45, 0x3ca2, 0x2c83, 0x1ce0, 0x0cc1,
			0xef1f, 0xff3e, 0xcf5d, 0xdf7c, 0xaf9b, 0xbfba, 0x8fd9, 0x9ff8,
			0x6e17, 0x7e36, 0x4e55, 0x5e74, 0x2e93, 0x3eb2, 0x0ed1, 0x1ef0
		]
		crc =  0xFFFF
		for i in range(0, len(data)):
			b = data[i] & 0xff
			b = (b ^ ((crc >> 8) & 0xff)) & 0xff
			crc = (TABLE_CRC16_CCITT[b] ^ (crc << 8)) & 0xffff
		return crc

	def exchange(self, cla, ins, p1, p2, data):
		apdu = bytearray([cla, ins, p1, p2, len(data)]) + bytearray(data)
		if self.card == None:
			print("%s" % binascii.hexlify(apdu))
		else:
			return self.card.exchange(apdu)

	def encryptAES(self, data):
		if not self.secure:
			return data
		paddedData = data + b'\x80'
		while (len(paddedData) % 16) != 0:
			paddedData += b'\x00'
		cipher = AES.new(self.key, AES.MODE_CBC, self.iv)
		encryptedData = cipher.encrypt(str(paddedData))
		self.iv = encryptedData[len(encryptedData) - 16:]
		return encryptedData

	def decryptAES(self, data):
		if not self.secure or len(data) == 0:
			return data
		cipher = AES.new(self.key, AES.MODE_CBC, self.iv)
		decryptedData = cipher.decrypt(data)
		l = len(decryptedData) - 1
		while (decryptedData[l] != chr(0x80)):
			l-=1
		decryptedData = decryptedData[0:l]
		self.iv = data[len(data) - 16:]
		return decryptedData

	def selectSegment(self, baseAddress):
		data = b'\x05' + struct.pack('>I', baseAddress)
		data = self.encryptAES(data)
		self.exchange(self.cla, 0x00, 0x00, 0x00, data)

	def loadSegmentChunk(self, offset, chunk):
		data = b'\x06' + struct.pack('>H', offset) + chunk
		data = self.encryptAES(data)
		self.exchange(self.cla, 0x00, 0x00, 0x00, data)		

	def flushSegment(self):
		data = b'\x07'
		data = self.encryptAES(data)
		self.exchange(self.cla, 0x00, 0x00, 0x00, data)				

	def crcSegment(self, offsetSegment, lengthSegment, crcExpected):
		data = b'\x08' + struct.pack('>H', offsetSegment) + struct.pack('>I', lengthSegment) + struct.pack('>H', crcExpected)
		data = self.encryptAES(data)
		self.exchange(self.cla, 0x00, 0x00, 0x00, data)						

	def validateTargetId(self, targetId):
		data = struct.pack('>I', targetId)
		self.exchange(self.cla, 0x04, 0x00, 0x00, data)

	def boot(self, bootadr, signature=None):
		# Force jump into Thumb mode
		bootadr |= 1
		data = b'\x09' + struct.pack('>I', bootadr)
		if (signature != None):
			data += chr(len(signature)) + signature
		data = self.encryptAES(data)
		self.exchange(self.cla, 0x00, 0x00, 0x00, data)

	def createApp(self, appflags, applength, appname, icon=None, path=None, iconOffset=None, iconSize=None, appversion=None):
		data = b'\x0B' + struct.pack('>I', applength) + struct.pack('>I', appflags) + struct.pack('>B', len(appname)) + appname
		if iconOffset is None:
			if not (icon is None):
				data += struct.pack('>B', len(icon)) + icon
			else:
				data += b'\x00'

		if not (path is None):
			data += struct.pack('>B', len(path)) + path
		else:
			data += b'\x00'

		if not iconOffset is None:
			data += struct.pack('>I', iconOffset) + struct.pack('>H', iconSize)

		if not appversion is None:
			data += struct.pack('>B', len(appversion)) + appversion

		data = self.encryptAES(data)
		self.exchange(self.cla, 0x00, 0x00, 0x00, data)						

	def deleteApp(self, appname):
		data = b'\x0C' +  struct.pack('>B',len(appname)) +  appname
		data = self.encryptAES(data)
		self.exchange(self.cla, 0x00, 0x00, 0x00, data)						

	def listApp(self, restart=True):
		if restart:
			data = b'\x0E'
		else:
			data = b'\x0F'
		data = self.encryptAES(data)
		response = str(self.exchange(self.cla, 0x00, 0x00, 0x00, data))
		response = bytearray(self.decryptAES(response))
		result = []
		offset = 0
		while offset != len(response):
			item = {}
			offset += 1
			item['name'] = response[offset + 1 : offset + 1 + response[offset]]
			offset += 1 + response[offset]
			item['flags'] = response[offset] << 24 | response[offset + 1] << 16 | response[offset + 2] << 8 | response[offset + 3]
			offset += 4
			item['hash'] = response[offset : offset + 32]
			offset += 32
			result.append(item)
		return result

	def load(self, erase_u8, max_length_per_apdu, hexFile):
		initialAddress = 0
		if self.relative:
			initialAddress = hexFile.minAddr()
		sha256 = hashlib.new('sha256')
		for area in hexFile.getAreas():
			startAddress = area.getStart() - initialAddress
			data = area.getData()
			self.selectSegment(startAddress)
			if len(data) == 0:
				continue
			if len(data) > 0x10000:
				raise Exception("Invalid data size for loader")
			crc = self.crc16(bytearray(data))
			offset = 0
			length = len(data)
			while (length > 0):
				if length > max_length_per_apdu - LOAD_SEGMENT_CHUNK_HEADER_LENGTH - MIN_PADDING_LENGTH:
					chunkLen = max_length_per_apdu - LOAD_SEGMENT_CHUNK_HEADER_LENGTH - MIN_PADDING_LENGTH
				else:
					chunkLen = length
				chunk = data[offset : offset + chunkLen]
				sha256.update(chunk)
				self.loadSegmentChunk(offset, bytes(chunk))
				offset += chunkLen
				length -= chunkLen
			self.flushSegment()
			self.crcSegment(0, len(data), crc)
		return sha256.hexdigest()

	def run(self, hexFile, bootaddr, signature=None):
		initialAddress = 0
		if self.relative:
			initialAddress = hexFile.minAddr()
		self.boot(bootaddr - initialAddress, signature)
		



"""
*******************************************************************************
*   Ledger Blue
*   (c) 2016 Ledger
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*      http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
********************************************************************************
"""

DEFAULT_ALIGNMENT = 1024

from .ecWrapper import PrivateKey
from .comm import getDongle
from .hexParser import IntelHexParser, IntelHexPrinter
from .hexLoader import HexLoader
from .deployed import getDeployedSecretV1, getDeployedSecretV2
import argparse
import struct
import binascii
import sys

def auto_int(x):
    return int(x, 0)

def parse_bip32_path(path, apilevel):
        if len(path) == 0:
                return b""
        result = b""
        elements = path.split('/')
        if apilevel >= 5:
        	result = result + chr(len(elements))
        for pathElement in elements:
                element = pathElement.split('\'')
                if len(element) == 1:
                        result = result + struct.pack(">I", int(element[0]))             
                else:
                        result = result + struct.pack(">I", 0x80000000 | int(element[0]))
        return result

parser = argparse.ArgumentParser()
parser.add_argument("--targetId", help="Set the chip target ID", type=auto_int)
parser.add_argument("--fileName", help="Set the file name to load")
parser.add_argument("--icon", help="Set the icon content to use (hex encoded)")
parser.add_argument("--curve", help="Curve on which the derivation is locked (secp256k1|prime256r1|ed25519), can be repeated", action='append')
parser.add_argument("--path", help="BIP 32 path to which the derivation is locked (format decimal a'/b'/c), can be repeated", action='append')
parser.add_argument("--appName", help="Set the application name")
parser.add_argument("--signature", help="Optional application's signature (hex encoded)")
parser.add_argument("--appFlags", help="Set the application flags", type=auto_int)
parser.add_argument("--bootAddr", help="Set the boot address", type=auto_int)
parser.add_argument("--rootPrivateKey", help="Set the root private key")
parser.add_argument("--apdu", help="Display APDU log", action='store_true')
parser.add_argument("--deployLegacy", help="Use legacy deployment API", action='store_true')
parser.add_argument("--apilevel", help="Use given API level when interacting with the device", type=auto_int)
parser.add_argument("--delete", help="Delete app before installing it", action='store_true')
parser.add_argument("--params", help="Store icon and install parameters in a parameter section before the code", action='store_true')
parser.add_argument("--appVersion", help="Set the application version (text)")

args = parser.parse_args()

if args.apilevel == None:
	args.apilevel = 5
if args.targetId == None:
	args.targetId = 0x31000002
if args.fileName == None:
	raise Exception("Missing fileName")
if args.appName == None:
	raise Exception("Missing appName")
if args.appFlags == None:
	args.appFlags = 0
if args.rootPrivateKey == None:
	privateKey = PrivateKey()
	publicKey = binascii.hexlify(privateKey.pubkey.serialize(compressed=False))
	print("Generated random root public key : %s" % publicKey)
	args.rootPrivateKey = privateKey.serialize()


if (sys.version_info.major == 3):
	args.appName = bytes(args.appName,'ascii')
if (sys.version_info.major == 2):
	args.appName = bytes(args.appName)

parser = IntelHexParser(args.fileName)
if args.bootAddr == None:
    args.bootAddr = parser.getBootAddr()

path = b""
curveMask = 0xff
if args.curve != None:	
	curveMask = 0x00
	for curve in args.curve:
		if curve == 'secp256k1':
			curveMask |= 0x01
		elif curve == 'prime256r1':
			curveMask |= 0x02
		elif curve == 'ed25519':
			curveMask |= 0x04
		else:
			raise Exception("Unknown curve " + curve)	

if args.apilevel >= 5:
	path += struct.pack('>B',curveMask)
	if args.path != None:
		for item in args.path:
			if len(item) != 0:
				path += parse_bip32_path(item, args.apilevel)	
else:
	if args.curve != None:
		print("Curve not supported using this API level, ignoring")
	if args.path != None:
		if len(args.path) > 1:
			print("Multiple path levels not supported using this API level, ignoring")
		else:
			path = parse_bip32_path(args.path[0], args.apilevel)	

icon = None
if not args.icon is None:
	icon = bytearray.fromhex(args.icon)
	
signature = None
if not args.signature is None:
	signature = bytearray.fromhex(args.signature)
	
#prepend app's data with the icon content (could also add other various install parameters)
printer = IntelHexPrinter(parser)
#todo build a TLV zone to keep install params
#todo dney nvm_write in that section ?
paramsSectionContent = []
if icon:
	paramsSectionContent = icon

# prepend the param section (arbitrary)
if (args.params):
	#take care of aligning the parameters sections to avoid possible invalid dereference of aligned words in the program nvram.
	#also use the default MPU alignment
	param_start = printer.minAddr()-len(paramsSectionContent)-(DEFAULT_ALIGNMENT-(len(paramsSectionContent)%DEFAULT_ALIGNMENT))
	printer.addArea(param_start, paramsSectionContent)

# account for added regions (install parameters, icon ...)
appLength = printer.maxAddr() - printer.minAddr()


dongle = getDongle(args.apdu)

if args.deployLegacy:
	secret = getDeployedSecretV1(dongle, bytearray.fromhex(args.rootPrivateKey), args.targetId)
else:
	secret = getDeployedSecretV2(dongle, bytearray.fromhex(args.rootPrivateKey), args.targetId)
loader = HexLoader(dongle, 0xe0, True, secret)

if (not (args.appFlags & 2)) and args.delete:
		loader.deleteApp(args.appName)

#heuristic to guess how to pass the icon
if (args.params):
	loader.createApp(args.appFlags, appLength, args.appName, None, path, 0, len(paramsSectionContent), args.appVersion)
else:
	loader.createApp(args.appFlags, appLength, args.appName, icon, path, None, None, args.appVersion)

hash = loader.load(0x0, 0xF0, printer)
print("Application hash : " + hash)
loader.run(printer, args.bootAddr, signature)



"""
*******************************************************************************
*   Ledger Blue
*   (c) 2016 Ledger
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*      http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
********************************************************************************
"""

from .comm import getDongle
import argparse
import getpass
import unicodedata
import sys

def auto_int(x):
    return int(x, 0)

parser = argparse.ArgumentParser()
parser.add_argument("--persistent", help="Persist passphrase as secondary PIN", action='store_true')

args = parser.parse_args()

dongle = getDongle(False)

passphrase = getpass.getpass("Enter BIP39 passphrase : ")
if isinstance(passphrase, bytes):
	passphrase = passphrase.decode(sys.stdin.encoding)
if len(passphrase) != 0:
	if args.persistent:
		p1 = 0x02
	else:
		p1 = 0x01
	passphrase = unicodedata.normalize('NFKD', passphrase)
	apdu = bytearray([0xE0, 0xD0, p1, 0x00, len(passphrase)]) + bytearray(passphrase, 'utf8')
	dongle.exchange(apdu, timeout=300)




"""
*******************************************************************************
*   Ledger Blue
*   (c) 2016 Ledger
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*      http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
********************************************************************************
"""

from abc import ABCMeta, abstractmethod
from .commException import CommException
from .ledgerWrapper import wrapCommandAPDU, unwrapResponseAPDU
from binascii import hexlify
import hid
import time
import sys

try:
	from smartcard.Exceptions import NoCardException
	from smartcard.System import readers
	from smartcard.util import toHexString, toBytes
	SCARD = True
except ImportError:
	SCARD = False

	
def hexstr(bstr):
	if (sys.version_info.major == 3):
		return hexlify(bstr).decode()
	if (sys.version_info.major == 2):
		return hexlify(bstr)
	return "<undecoded APDU<"

class DongleWait(object):
	__metaclass__ = ABCMeta

	@abstractmethod
	def waitFirstResponse(self, timeout):
		pass

class Dongle(object):
	__metaclass__ = ABCMeta

	@abstractmethod
	def exchange(self, apdu, timeout=20):
		pass

	@abstractmethod
	def close(self):
		pass

	def setWaitImpl(self, waitImpl):
		self.waitImpl = waitImpl

class HIDDongleHIDAPI(Dongle, DongleWait):

	def __init__(self, device, ledger=False, debug=False):
		self.device = device
		self.ledger = ledger		
		self.debug = debug
		self.waitImpl = self
		self.opened = True

	def exchange(self, apdu, timeout=20):
		if self.debug:
			print("=> %s" % hexstr(apdu))
		if self.ledger:
			apdu = wrapCommandAPDU(0x0101, apdu, 64)		
		padSize = len(apdu) % 64
		tmp = apdu
		if padSize != 0:
			tmp.extend([0] * (64 - padSize))
		offset = 0
		while(offset != len(tmp)):
			data = tmp[offset:offset + 64]
			data = bytearray([0]) + data
			self.device.write(data)
			offset += 64
		dataLength = 0
		dataStart = 2		
		result = self.waitImpl.waitFirstResponse(timeout)
		if not self.ledger:
			if result[0] == 0x61: # 61xx : data available
				self.device.set_nonblocking(False)
				dataLength = result[1]
				dataLength += 2
				if dataLength > 62:
					remaining = dataLength - 62
					while(remaining != 0):
						if remaining > 64:
							blockLength = 64
						else:
							blockLength = remaining
						result.extend(bytearray(self.device.read(65))[0:blockLength])
						remaining -= blockLength
				swOffset = dataLength
				dataLength -= 2
				self.device.set_nonblocking(True)
			else:
				swOffset = 0
		else:
			self.device.set_nonblocking(False)
			while True:
				response = unwrapResponseAPDU(0x0101, result, 64)
				if response is not None:
					result = response
					dataStart = 0
					swOffset = len(response) - 2
					dataLength = len(response) - 2
					self.device.set_nonblocking(True)
					break
				result.extend(bytearray(self.device.read(65)))
		sw = (result[swOffset] << 8) + result[swOffset + 1]
		response = result[dataStart : dataLength + dataStart]
		if self.debug:
			print("<= %s%.2x" % (hexstr(response), sw))
		if sw != 0x9000:
			raise CommException("Invalid status %04x" % sw, sw, response)
		return response

	def waitFirstResponse(self, timeout):
		start = time.time()
		data = ""
		while len(data) == 0:
			data = self.device.read(65)
			if not len(data):
				if time.time() - start > timeout:
					raise CommException("Timeout")
				time.sleep(0.0001)
		return bytearray(data)

	def close(self):
		if self.opened:
			try:
				self.device.close()
			except:
				pass
		self.opened = False

class DongleSmartcard(Dongle):

	def __init__(self, device, debug=False):
		self.device = device
		self.debug = debug
		self.waitImpl = self
		self.opened = True

	def exchange(self, apdu, timeout=20):
		if self.debug:
			print("=> %s" % hexstr(apdu))
		response, sw1, sw2 = self.device.transmit(toBytes(hexlify(apdu)))
		sw = (sw1 << 8) | sw2
		if self.debug:
			print("<= %s%.2x" % (hexstr(response).replace(" ", ""), sw))
		if sw != 0x9000:
			raise CommException("Invalid status %04x" % sw, sw, bytearray(response))
		return bytearray(response)

	def close(self):
		if self.opened:
			try:
				self.device.disconnect()
			except:
				pass
		self.opened = False

def getDongle(debug=False, selectCommand=None):
	dev = None
	hidDevicePath = None
	ledger = True
	for hidDevice in hid.enumerate(0, 0):
		if hidDevice['vendor_id'] == 0x2c97:
			hidDevicePath = hidDevice['path']
	if hidDevicePath is not None:
		dev = hid.device()
		dev.open_path(hidDevicePath)
		dev.set_nonblocking(True)
		return HIDDongleHIDAPI(dev, ledger, debug)
	if SCARD:
		connection = None
		for reader in readers():
			try:
				connection = reader.createConnection()
				connection.connect()				
				if selectCommand != None:
					response, sw1, sw2 = connection.transmit(toBytes("00A4040010FF4C4547522E57414C5430312E493031"))																  
					sw = (sw1 << 8) | sw2
					if sw == 0x9000:
						break
					else:
						connection.disconnect()
						connection = None
				else:
					break
			except:
				connection = None
				pass
		if connection is not None:
			return DongleSmartcard(connection, debug)
	raise CommException("No dongle found")




"""
*******************************************************************************
*   Ledger Blue
*   (c) 2016 Ledger
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*      http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
********************************************************************************
"""

from .comm import getDongle
from .deployed import getDeployedSecretV2
from .ecWrapper import PrivateKey
from Crypto.Cipher import AES
import argparse
import sys
import fileinput


def auto_int(x):
    return int(x, 0)

parser = argparse.ArgumentParser()
parser.add_argument("--fileName", help="Set the file name to load")
parser.add_argument("--apdu", help="Display APDU log", action='store_true')
parser.add_argument(
    "--scp", help="open secure channel to exchange apdu", action='store_true')
parser.add_argument("--targetId", help="Set the chip target ID", type=auto_int)
parser.add_argument("--rootPrivateKey", help="Set the root private key")

args = parser.parse_args()

if args.targetId is None:
    args.targetId = 0x31100002
if not args.fileName:
    #raise Exception("Missing fileName")
    file = sys.stdin
else:
    file = open(args.fileName, "r")


class SCP:

    def __init__(self, dongle, targetId, rootPrivateKey):
        self.key = getDeployedSecretV2(dongle, rootPrivateKey, targetId)
        self.iv = b'\x00' * 16

    def encryptAES(self, data):
        paddedData = data + b'\x80'
        while (len(paddedData) % 16) != 0:
            paddedData += b'\x00'
        cipher = AES.new(self.key, AES.MODE_CBC, self.iv)
        encryptedData = cipher.encrypt(paddedData)
        self.iv = encryptedData[len(encryptedData) - 16:]
        return encryptedData

    def decryptAES(self, data):
        if len(data) == 0:
            return data
        cipher = AES.new(self.key, AES.MODE_CBC, self.iv)
        decryptedData = cipher.decrypt(data)
        l = len(decryptedData) - 1
        while (decryptedData[l] != chr(0x80)):
            l -= 1
        decryptedData = decryptedData[0:l]
        self.iv = data[len(data) - 16:]
        return decryptedData

dongle = getDongle(args.apdu)
if args.scp:
    if args.rootPrivateKey is None:
        privateKey = PrivateKey()
        publicKey = binascii.hexlify(
            privateKey.pubkey.serialize(compressed=False))
        print("Generated random root public key : %s" % publicKey)
        args.rootPrivateKey = privateKey.serialize()
    scp = SCP(dongle, args.targetId, bytearray.fromhex(args.rootPrivateKey))

for data in file:
    data = data.rstrip('\r\n').decode('hex')
    if len(data) < 5:
        continue
    if args.scp:
        data = bytearray(data)
        if data[4] > 0 and len(data) > 5:
            apduData = data[5: 5 + data[4]]
            apduData = scp.encryptAES(str(apduData))
            result = dongle.exchange(
                data[0:4] + bytearray([len(apduData)]) + bytearray(apduData))
        else:
            result = dongle.exchange(data[0:5])
        result = scp.decryptAES(str(result))
        if args.apdu:
            print("<= Clear " + result.encode('hex'))
    else:
        dongle.exchange(bytearray(data))
